#include <torch/extension.h>
#include <ATen/ATen.h>
#include <cuda.h>
#include <cuda_runtime.h>

// Optimized kernel to compute the mean of each (batch, channel) slice and accumulate the result
// into a global accumulator to compute the overall mean in one pass.

template <unsigned int blockSize>
__global__ void optimized_mean_kernel(
    const float* __restrict__ input,
    float* __restrict__ global_accum,
    int H,
    int W,
    int C
) {
    extern __shared__ float shared[];  // shared memory for block reduction
    int num_elements = H * W;
    int batch = blockIdx.x / C;
    int channel = blockIdx.x % C;
    int input_offset = (batch * C + channel) * num_elements;

    float sum = 0.0f;
    // Each thread accumulates partial sum over its strided elements
    for (int i = threadIdx.x; i < num_elements; i += blockSize) {
        sum += input[input_offset + i];
    }

    // Store partial sums into shared memory
    shared[threadIdx.x] = sum;
    __syncthreads();

    // Intra-block reduction using sequential addressing reduction
    for (unsigned int s = blockSize / 2; s > 0; s >>= 1) {
        if (threadIdx.x < s) {
            shared[threadIdx.x] += shared[threadIdx.x + s];
        }
        __syncthreads();
    }

    // Thread 0 computes the channel mean and atomically adds it to the global accumulator
    if (threadIdx.x == 0) {
        float channel_mean = shared[0] / static_cast<float>(num_elements);
        atomicAdd(global_accum, channel_mean);
    }
}

at::Tensor module_fn(
    at::Tensor x,
    int64_t stride,
    int64_t padding,
    int64_t output_padding,
    at::Tensor conv_transpose,
    at::Tensor conv_transpose_bias,
    double multiplier
) {
    // Perform transposed convolution using PyTorch's native function
    at::Tensor y = at::conv_transpose2d(
        x,
        conv_transpose,
        conv_transpose_bias,
        {stride, stride},
        {padding, padding},
        {output_padding, output_padding},
        1,
        {1, 1}
    );

    // Scale the output by the multiplier
    y = y * multiplier;

    // Get dimensions of the result tensor
    auto dims = y.sizes();
    int N = dims[0];
    int C = dims[1];
    int H = dims[2];
    int W = dims[3];

    // Allocate a scalar accumulator on device and initialize to zero
    auto options = torch::TensorOptions().device(y.device()).dtype(y.dtype());
    at::Tensor accum = torch::zeros({1}, options);

    // Launch one block per (batch, channel) pair
    constexpr int blockSize = 256;
    int numBlocks = N * C;
    size_t sharedMemSize = blockSize * sizeof(float);

    optimized_mean_kernel<blockSize><<<numBlocks, blockSize, sharedMemSize>>>(
        y.data_ptr<float>(),
        accum.data_ptr<float>(),
        H, W, C
    );

    // Compute final mean by dividing the accumulated channel means by total number of (batch, channel) pairs
    accum = accum / static_cast<float>(N * C);
    return accum;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &module_fn, "Module function");
}
